#include <stdlib.h>

#include <arch.h>
#include <device.h>
#include <common.h>
#include <error.h>
#include <dbgmem.h>
#include <zobject.h>
#include "render.h"

signed_32bit *scanlines=0; /* Contains startx,endx for all scanlines */
signed_32bit *scanlinedata=0; /* Contains other data for lines (u,v,c..) */
signed_32bit GradientX[MAXDATADIM];
signed_32bit GradientY[MAXDATADIM];
signed_32bit TopY,BottomY;

byte *ColorTransform;

int InitRenderer(void)
{
  int fd;

  ColorTransform=malloc(16*16);
  if (ColorTransform==0) {
    return 0;
  }
  fd=open("standard.tfm",OPEN_READ_BINARY);
  if (fd==0)
    return 0;
  read(fd,ColorTransform,16*16);
  close(fd);
  return 1;
}

void RenderZObject(ZObject *ZObj)
{
  int i;

  if (scanlines==0) {
    scanlines=malloc(sizeof(signed_32bit)*(HW_ScreenSizeY+1)*2);
    if (scanlines==0)
      ErrError("Out of memory.");
    for (i=0;i<(HW_ScreenSizeY+1)*2;i++) {
      scanlines[i]=0;
    }
  }
  if (scanlinedata==0) {
    scanlinedata=malloc(sizeof(signed_32bit)*HW_ScreenSizeY*MAXDATADIM);
    if (scanlinedata==0)
      ErrError("Out of memory.");
  }

  switch (ZObj->ObjType) {
  case TYPE_ZFLATPOLY:
    RenderFlatPoly((ZFlatPoly *)ZObj);
    break;
  case TYPE_ZGOROUDPOLY:
    RenderGoroudPoly((ZGoroudPoly *)ZObj);
    break;
  case TYPE_ZTEXTUREPOLY:
    RenderTexturePoly((ZTexturePoly *)ZObj);
    break;
  case TYPE_ZLINE:
    RenderLine((ZLine *)ZObj);
    break;
  case TYPE_ZDOT:
    RenderDot((ZDot *)ZObj);
    break;
  default:
    break;
  }
}

void RenderDot(ZDot *dot)
{
  signed_32bit x,y;

  x=(dot->Vertices[0]+8)>>4;
  y=(dot->Vertices[1]+8)>>4;

  if (x<0)
    return;
  if (x>=HW_ScreenSizeX)
    return;
  if (y<0)
    return;
  if (y>=HW_ScreenSizeY)
    return;

  *(HW_OffscreenBuffer+x+y*HW_BytesPerRow)=dot->Color;
}

void RenderLine(ZLine *line)
{
  signed_32bit dx,dy;

  dx=line->Vertices[2]-line->Vertices[0];
  dy=line->Vertices[3]-line->Vertices[1];
}


void RenderFlatPoly(ZFlatPoly *poly)
{
  int i;
  byte *buffptr;
  int startx,endx;
  int color;

  RenderPoly(&(poly->Vertices[0]),2,poly->ClockWise);

  color=ColorTransform[poly->Color*16+(poly->Intensity>>(16-4))];

  buffptr=HW_OffscreenBuffer+TopY*HW_BytesPerRow;
  for (i=TopY*2;i<=BottomY*2;i+=2) {
    startx=scanlines[i];
    if (startx<0)
	startx=0;
    endx=scanlines[i+1];
    if (endx>HW_ScreenSizeX)
	endx=HW_ScreenSizeX;
    if (startx<endx) {
      HW_MemSet(buffptr+startx,color,
		endx-startx);
      scanlines[i]=0;
      scanlines[i+1]=0;
    }
    buffptr+=HW_BytesPerRow;
  }
}

#ifdef ARCH_HAS_GOROUD_H
#include <goroud.h>
#else
void RenderGoroudLine(byte *ptr,int len,
		      signed_32bit Int,signed_32bit dInt,
		      byte *coltfm )
{
  byte *toptr;
  toptr=ptr+(len&0xffffc);
  len-=len&0xfffffc;
  while(toptr!=ptr) {
    *ptr++=coltfm[Int>>(16-4)];
    Int+=dInt;
    *ptr++=coltfm[Int>>(16-4)];
    Int+=dInt;
    *ptr++=coltfm[Int>>(16-4)];
    Int+=dInt;
    *ptr++=coltfm[Int>>(16-4)];
    Int+=dInt;
  }
  while (len-->0) {
    *ptr++=coltfm[Int>>(16-4)];
    Int+=dInt;
  }
}
#endif
void RenderGoroudPoly(ZGoroudPoly *poly)
{
  int i;
  byte *buffptr;
  signed_32bit dInt;
  int startx,endx;
  byte *coltfm;

  coltfm=&ColorTransform[poly->Color*16];

  RenderPoly(&(poly->Vertices[0]),3,poly->ClockWise);

  dInt=GradientX[0];
  buffptr=HW_OffscreenBuffer+TopY*HW_BytesPerRow;
  for (i=TopY*2;i<=BottomY*2;i+=2) {
    startx=scanlines[i];
    if (startx<0)
	startx=0;
    endx=scanlines[i+1];
    if (endx>HW_ScreenSizeX)
	endx=HW_ScreenSizeX;
    if (startx<endx) {
      /*
      if ((scanlinedata[(i>>1)*MAXDATADIM]+dInt*(startx-scanlines[i]))<0) {
	printf("bug!!!!!!!!!!!!!!!!!!!!!!!!\n");
      }
      */
      RenderGoroudLine(buffptr+startx,endx-startx,
	  scanlinedata[(i>>1)*MAXDATADIM]+dInt*(startx-scanlines[i]),
	  dInt,coltfm);
      scanlines[i]=0;
      scanlines[i+1]=0;
    }
    buffptr+=HW_BytesPerRow;
  }
}


void RenderTexturePoly(ZTexturePoly *poly)
{
  int i;
  byte *buffptr;
  byte *ptr,*toptr;
  byte *texture;
  byte bitwidth;
  int len;
  register signed_32bit U,V;
  signed_32bit dU,dV;
  int tmppos;
  int startx,endx,skipx;

  RenderPoly(&(poly->Vertices[0]),4,poly->ClockWise);

  texture=poly->Texture;
  bitwidth=poly->BitWidth;

  buffptr=HW_OffscreenBuffer+TopY*HW_BytesPerRow;
  dU=GradientX[0];
  dV=GradientX[1];
  for (i=TopY*2;i<=BottomY*2;i+=2) {
    startx=scanlines[i];
    if (startx<0)
	startx=0;
    endx=scanlines[i+1];
    if (endx>HW_ScreenSizeX)
	endx=HW_ScreenSizeX;
    if ( (len=endx-startx) > 0 ) {
      ptr=buffptr+startx;
      skipx=startx-scanlines[i];
      tmppos=(i>>1)*MAXDATADIM;
      U=scanlinedata[tmppos];
      V=scanlinedata[tmppos+1];
      if (skipx) {
	U+=dU*skipx;
	V+=dV*skipx;
      }
      toptr=ptr+(len&0xfffffe);
      while(ptr!=toptr) {
	*ptr++=texture[(U>>16)+((V>>16)<<bitwidth)];
	U+=dU;
	V+=dV;
	*ptr++=texture[(U>>16)+((V>>16)<<bitwidth)];
	U+=dU;
	V+=dV;
      }
      if (len&1) {
	*ptr=texture[(U>>16)+((V>>16)<<bitwidth)];
      }
/* not unrolled inner loop:*/
/*      while(len--) {
	u=U>>16;
	v=V>>16;
	*ptr++=texture[u+(v<<bitwidth)];
	U+=dU;
	V+=dV;
      }*/

      scanlines[i]=0;
      scanlines[i+1]=0;
    }
    buffptr+=HW_BytesPerRow;
  }
}

signed_32bit Ceil28_4(signed_32bit fract)
{
  signed_32bit ret;
  signed_32bit num;
  
  num=fract-1+16;
  if (num>=0) {
    return num>>4;
  } else {
    ret=-((-num)>>4);
    ret-=((-num)&15)?1:0;
    return ret;
  }  
}

signed_32bit Mult28_4(signed_32bit a,signed_32bit b)
{
  return (a*b)>>4;
}

void FloorDivMod(signed_32bit numerator,signed_32bit denominator,
		 signed_32bit *floor,signed_32bit *mod)
{
  if (numerator>=0) {
    *floor=numerator/denominator;
    *mod=numerator%denominator;
  } else {
    *floor=-((-numerator)/denominator);
    *mod=(-numerator)%denominator;
    if (*mod) {
      (*floor)--;
      *mod=denominator-*mod;
    }
  }
}


void TraceEdge(signed_32bit *start,signed_32bit *end,int right,int datadim)
{
  int i;
  signed_32bit y0,y1,skipy,skipx;
  signed_32bit height;
  signed_32bit dN,dM,X,Y,R;
  signed_32bit XStep,ErrorTerm,ErrorTermAdvance,Denominator;
  signed_32bit data[MAXDATADIM];
  signed_32bit datainc[MAXDATADIM];
  signed_32bit XPreStep,YPreStep;

  signed_32bit Xloop,ErrorTermloop,heightloop;
  signed_32bit *scanlineptr;
  

  y0=Ceil28_4(start[1]);
  skipy=y0;
  if (y0<0)
    y0=0;
  skipy=y0-skipy;
  skipx=Ceil28_4(start[0]);

  y1=Ceil28_4(end[1]);
  if (y1>HW_ScreenSizeY)
    y1=HW_ScreenSizeY;

  height=y1-y0;
  if (height>0) {
    dN=end[1]-start[1];
    dM=end[0]-start[0];
    R=dM*16*y0-dM*start[1]+dN*start[0]-1+dN*16;
    Denominator=16*dN;
    
    FloorDivMod(16*dM,Denominator,&XStep,&ErrorTermAdvance);
    FloorDivMod(R,Denominator,&X,&ErrorTerm);
    skipx=X-skipx;

    if (datadim>0) {
      XPreStep=(X*16-start[0])<<(16-4);
      YPreStep=(y0*16-start[1])<<(16-4);

      for (i=0;i<datadim;i++) {
	data[i]=start[2+i]+FpMult(GradientY[i],YPreStep)
	  +FpMult(GradientX[i],XPreStep);
	datainc[i]=XStep*GradientX[i]+GradientY[i];
      }
    }


    Y=y0;
    if (right || (datadim==0)) {
      scanlineptr=&scanlines[Y*2+right];
      Xloop=X;
      ErrorTermloop=ErrorTerm;
      heightloop=height;
      while(heightloop--) {
	*scanlineptr=Xloop;
	Xloop+=XStep;
	scanlineptr+=2;
	ErrorTermloop+=ErrorTermAdvance;
	if (ErrorTermloop>=Denominator) {
	  Xloop++;
	  ErrorTermloop-=Denominator;
	}
      }
    } else if (datadim==1) {
      heightloop=height;
      Xloop=X;
      while(heightloop--) {
	/*	if (data[0]<0)
	  data[0]=0;*/ /* ugly fix */
	scanlines[Y*2]=Xloop;  /* right is always 0 here */
	scanlinedata[Y*MAXDATADIM]=data[0];
	data[0]+=datainc[0];
	Xloop+=XStep;
	Y++;
	ErrorTerm+=ErrorTermAdvance;
	if (ErrorTerm>=Denominator) {
	  data[0]+=GradientX[0];
	  Xloop++;
	  ErrorTerm-=Denominator;
	}
      }
    } else if (datadim==2) {
      heightloop=height;
      Xloop=X;
      while(heightloop--) {
	/*if (data[0]<0)
	  data[0]=0; *//* ugly fix */
	/*if (data[1]<0)
	  data[1]=0; *//* ugly fix */
	scanlines[Y*2]=Xloop;  /* right is always 0 here */
	scanlinedata[Y*MAXDATADIM]=data[0];
	data[0]+=datainc[0];
	scanlinedata[Y*MAXDATADIM+1]=data[1];
	data[1]+=datainc[1];
	Xloop+=XStep;
	Y++;
	ErrorTerm+=ErrorTermAdvance;
	if (ErrorTerm>=Denominator) {
	  data[0]+=GradientX[0];
	  data[1]+=GradientX[1];
	  Xloop++;
	  ErrorTerm-=Denominator;
	}
      }
    } else {
      assert(0/*Shouldn't come here*/);
    }
  }
}


void CalculateGradients(signed_32bit *poly,int dim)
{
  int i;

  float dc1,dc2,dx1,dx2,dy1,dy2,denom;

  /* Calcluate gradients for interpolating data: */

  dx1=(poly[1*dim]-poly[2*dim]);
  dx2=(poly[0*dim]-poly[2*dim]);
  dy1=(poly[1*dim+1]-poly[2*dim+1]);
  dy2=(poly[0*dim+1]-poly[2*dim+1]); 
  denom=(dx1*dy2-dx2*dy1)/16;
  if ((denom>0.01)||(denom<-0.01)) {
    for (i=0;i<(dim-2);i++) {
      dc1=(poly[1*dim+2+i]-poly[2*dim+2+i]); 
      dc2=(poly[0*dim+2+i]-poly[2*dim+2+i]);
      GradientX[i]=(dc1*dy2-dc2*dy1)/(denom);
      GradientY[i]=(dc1*dx2-dc2*dx1)/(-denom);
    }
  } else {
    for (i=0;i<(dim-2);i++) {
      GradientX[i]=0;
      GradientY[i]=0;
    }
  }
}



void RenderPoly(signed_32bit *poly,int dim,byte ClockWise)
{
  int i,j,left;
  int TopVertex,BottomVertex;

  if (dim>2)
    CalculateGradients(poly,dim);

  /* Find Top Vertex: */
  BottomVertex=TopVertex=0;
  BottomY=TopY=poly[1];
  for (i=1;i<3;i++) {
    int tmpy=poly[dim*i+1];
    if (tmpy < TopY) {
      TopVertex=i;
      TopY=tmpy;
    }
    if (tmpy > BottomY) {
      BottomVertex=i;
      BottomY=tmpy;
    }
  }
  TopY=Ceil28_4(TopY);
  BottomY=Ceil28_4(BottomY);
  if (TopY<0) {
    TopY=0;
  }
  if (BottomY<0) {
    TopY=0;
    BottomY=0;
    return;  /* Clip whole poly */
  }
  if (TopY>=HW_ScreenSizeY) {
    TopY=0;
    BottomY=0;
    return;  /* Clip whole poly */
  }
  if (BottomY>=HW_ScreenSizeY) {
    BottomY=HW_ScreenSizeY-1;
  }

  if (ClockWise)
    left=-1;
  else
    left=1;

  i=TopVertex;
  j=i+left;
  if (j<0) j=2;
  if (j>2) j=0;
   while(i!=BottomVertex) { 
    TraceEdge(&poly[i*dim],&poly[j*dim],0/*left*/,dim-2);
    i=j;
    j=i+left;
    if (j<0) j=2;
    if (j>2) j=0;
  }

  i=TopVertex;
  j=i-left;
  if (j<0) j=2;
  if (j>2) j=0;
  while(i!=BottomVertex) { 
    TraceEdge(&poly[i*dim],&poly[j*dim],1/*right*/,dim-2);
    i=j;
    j=i-left;
    if (j<0) j=2;
    if (j>2) j=0;
  }
}
