#ifndef VECTOR_H
#define VECTOR_H

#include <arch.h>

typedef real matrix[3][4];
typedef struct vector {
  real x;
  real y;
  real z;
} vector;

/* General defines: */
#define UNITMATRIX {{1.0,0,0,0},{0,1.0,0,0},{0,0,1.0,0}}

/* vectors and matrix functions */
extern void MCopy(matrix source,matrix dest);
extern int MCompare(matrix a, matrix b);
extern int VCompare(vector *a, vector *b);
extern void MMult(matrix a,matrix b,matrix c);
extern INLINE void VMult4(matrix a,vector *b,vector *c)
#ifdef HAS_INLINE
{
  c->x= a[0][0]*b->x + a[0][1]*b->y + a[0][2]*b->z + a[0][3];
  c->y= a[1][0]*b->x + a[1][1]*b->y + a[1][2]*b->z + a[1][3];
  c->z= a[2][0]*b->x + a[2][1]*b->y + a[2][2]*b->z + a[2][3];
}
#endif
;

extern INLINE void VMult3(matrix a,vector *b,vector *c)
#ifdef HAS_INLINE
{
  c->x= a[0][0]*b->x + a[0][1]*b->y + a[0][2]*b->z;
  c->y= a[1][0]*b->x + a[1][1]*b->y + a[1][2]*b->z;
  c->z= a[2][0]*b->x + a[2][1]*b->y + a[2][2]*b->z;
}
#endif
;

extern void MRotX(real theta,matrix dest);
extern void MRotY(real theta,matrix dest);
extern void MRotZ(real theta,matrix dest);
extern INLINE real VDot(vector *a,vector *b)
#ifdef HAS_INLINE
{
  return a->x*b->x + a->y*b->y + a->z*b->z;
}
#endif
;

extern void VCross(vector *a,vector *b,vector *c);
extern void MOrtoNormalize(matrix a);
extern real DetM(matrix a);
extern void InvertM(matrix a);

#endif /* VECTOR_H */
